# -*- coding: utf-8 -*-
import json
from datetime import datetime
from unittest.mock import Mock

import pytest
from django.core.exceptions import ValidationError
from django.db.models import Count
from django.test import TestCase
from django.urls import resolve, reverse
from djiffy.models import Canvas, Manifest

from derrida.books.models import (CreatorType, DerridaWork, DerridaWorkSection,
                                  Instance, InstanceCatalogue, InstanceCreator,
                                  InstanceLanguage, Journal, Language,
                                  OwningInstitution, Publisher, Reference,
                                  ReferenceType, Subject, Work, WorkLanguage,
                                  WorkSubject)
from derrida.interventions.models import Intervention
from derrida.people.models import Person
from derrida.places.models import Place


class TestOwningInstitution(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        long_name = 'New York Society Library'
        short_name = 'NYSL'
        inst = OwningInstitution(name=long_name)
        # should use long name if no short name is set
        assert str(inst) == long_name
        inst.short_name = short_name
        assert str(inst) == short_name

    def test_instance_count(self):
        # test abstract book count mix-in via owning institution model
        # tests that html for admin form is rendered correctly

        pl = Place.objects.first()
        inst = OwningInstitution.objects.create(name='NYSL',
            place=pl)
        # new institution has no books associated
        change_url = reverse('admin:books_instance_changelist')
        admin_book_count = inst.instance_count()
        assert change_url in admin_book_count
        assert 'id__exact=%s' % inst.pk in admin_book_count
        assert '0' in admin_book_count

        # create a book and associated it with the institution
        pub = Publisher.objects.create(name='Pub Lee')
        wk = Work.objects.create(primary_title='Some title')
        instance = Instance.objects.create(work=wk)
         # publisher=pub, pub_place=pl,
            # is_extant=False, is_annotated=False, is_digitized=False)

        cat = InstanceCatalogue.objects.create(institution=inst,
            instance=instance, is_current=False)

        inst_book_count = inst.instance_count()
        assert change_url in inst_book_count
        assert 'id__exact=%s' % inst.pk in inst_book_count
        assert '1' in inst_book_count


class TestDerridaWork(TestCase):

    def setUp(self):
        testwork, c = DerridaWork.objects.get_or_create(
            short_title='Ceci n\'est pas un livre',
            full_citation=('Ceci n\'est pas un livre: '
                           'and other tales of deconstructionism'),
            is_primary=True,
        )

    def test_str(self):
        '''Test that DerridaWork produces its expected string'''
        short_title = 'Ceci n\'est pas un livre'
        testwork = DerridaWork.objects.get(short_title=short_title)
        assert str(testwork) == short_title


class TestDerridaSection(TestCase):

    def test_str(self):
        assert str(DerridaWorkSection(name='Chapter 1')) == 'Chapter 1'


class TestReference(TestCase):
    fixtures = ['sample_work_data']

    def setUp(self):
        self.manif = Manifest.objects.create()
        self.la_vie = Instance.objects.get(work__short_title__contains="La vie")
        self.dg = DerridaWork.objects.get(pk=1)
        self.quotation = ReferenceType.objects.get(name='Quotation')

    def test_str(self):
        # Writing this out because complicated output
        desired_output = 'De la grammatologie, 110a: %s, 10s, Quotation' % \
            self.la_vie.display_title()
        reference = Reference.objects.create(
            instance=self.la_vie,
            derridawork=self.dg,
            derridawork_page='110',
            derridawork_pageloc='a',
            book_page='10s',
            reference_type=self.quotation
        )
        assert str(reference) == desired_output

    def test_get_absolute_url(self):
        ref = Reference.objects.create(
            instance=self.la_vie,
            derridawork=self.dg,
            derridawork_page='110',
            derridawork_pageloc='a',
            book_page='10s',
            reference_type=self.quotation
        )
        ref_url = ref.get_absolute_url()
        resolved_url = resolve(ref_url)
        assert resolved_url.url_name == 'reference'
        assert resolved_url.namespace == 'books'
        assert resolved_url.kwargs['derridawork_slug'] == self.dg.slug
        assert resolved_url.kwargs['page'] == ref.derridawork_page
        assert resolved_url.kwargs['pageloc'] == ref.derridawork_pageloc

    def test_instance_ids_with_digital_editions(self):
        # check static method, so we don't need cls or self
        Reference.instance_ids_with_digital_editions()

        # no instances have associated canvases so this should return an
        # empty list as a JSON string
        data = Reference.instance_ids_with_digital_editions()
        assert json.loads(data) == []

        # add a canvas to la_vie, then it should appear in the list
        self.la_vie.digital_edition = self.manif
        self.la_vie.save()
        data = Reference.instance_ids_with_digital_editions()
        assert json.loads(data) == [self.la_vie.pk]

    def test_instance_slug(self):
        # create a reference
        ref = Reference.objects.create(
            instance=self.la_vie,
            derridawork=self.dg,
            derridawork_page='110',
            derridawork_pageloc='a',
            book_page='10s',
            reference_type=self.quotation
        )
        # not a book section (none in test set are)
        # should return the slug of its instance
        assert ref.instance_slug == self.la_vie.slug

        # make work into a book section as a 'collected in'
        la_vie_collected = Instance.objects.create(work=self.la_vie.work,
            slug='la-vie-collected')
        self.la_vie.collected_in = la_vie_collected
        self.la_vie.save()
        # should return the slug for the collection
        assert ref.instance_slug == la_vie_collected.slug

    def test_get_instance_url(self):
        # create a reference
        ref = Reference.objects.create(
            instance=self.la_vie,
            derridawork=self.dg,
            derridawork_page='110',
            derridawork_pageloc='a',
            book_page='10s',
            reference_type=self.quotation
        )
        # not a book section (none in test set are)
        # should return the slug of its instance
        assert ref.instance_url == self.la_vie.get_absolute_url()

        # make work into a book section as a 'collected in'
        la_vie_collected = Instance.objects.create(work=self.la_vie.work,
            slug='la-vie-collected')
        self.la_vie.collected_in = la_vie_collected
        self.la_vie.save()
        # should return the slug for the collection
        assert ref.instance_url == la_vie_collected.get_absolute_url()

    def test_book(self):
        ref = Reference.objects.create(
            instance=self.la_vie,
            derridawork=self.dg,
            derridawork_page='110',
            derridawork_pageloc='a',
            reference_type=self.quotation
        )
        # not a book section (none in test set are)
        # should return the instance
        assert ref.book == self.la_vie

        # create a book section and reassociated the reference
        vie_part_wk = Work.objects.create()
        vie_part = Instance.objects.create(work=vie_part_wk,
            collected_in=self.la_vie)
        ref.instance = vie_part
        # book should return the collected work
        assert ref.book == self.la_vie


class TestReferenceQuerySet(TestCase):
    fixtures = ['test_references.json']

    def setUp(self):
        self.ref_qs = Reference.objects.all()

    def test_order_by_source_page(self):
        pages = sorted([ref.derridawork_page for ref in self.ref_qs.all()])
        assert list(Reference.objects.order_by_source_page()
                             .values_list('derridawork_page', flat=True)) \
                == pages

    def test_order_by_author(self):
        authors = sorted(
            ['; '.join([str(p) for p in ref.instance.work.authors.all()])
             for ref in self.ref_qs])
        qs_authors = list(self.ref_qs.order_by_author() \
           .values_list('instance__work__authors__authorized_name', flat=True))
        qs_authors = ['' if name is None else name for name in qs_authors]
        assert qs_authors == authors

    def test_summary_values(self):
        ref = self.ref_qs.first()
        ref_values = self.ref_qs.summary_values().first()
        assert ref_values['id'] == ref.pk
        assert ref_values['instance__slug'] == ref.instance.slug
        assert ref_values['derridawork__slug'] == ref.derridawork.slug
        assert ref_values['derridawork_page'] == ref.derridawork_page
        assert ref_values['derridawork_pageloc'] == ref.derridawork_pageloc
        # no author by default
        assert 'author' not in ref_values

        # author should be included if requested
        ref_values = self.ref_qs.summary_values(include_author=True).first()
        assert ref_values['author'] == \
            ref.instance.work.authors.first().authorized_name


class TestWork(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        la_vie = Work.objects.get(short_title__contains="La vie")
        assert '%s (%s)' % (la_vie.short_title, la_vie.year) \
            == str(la_vie)
        # no date
        la_vie.year = None
        assert '%s (n.d.)' % (la_vie.short_title, )

    def test_author_names(self):
        la_vie = Work.objects.get(short_title__contains="La vie")
        assert la_vie.author_names() == "L\u00e9vi-Strauss"

    def test_instance_count(self):
        la_vie = Work.objects.get(short_title__contains="La vie")
        inst_count = la_vie.instance_count()
        instance_list_url = reverse('admin:books_instance_changelist')
        assert instance_list_url in inst_count
        assert 'id__exact=%s' % la_vie.pk in inst_count
        assert '1' in inst_count


class TestInstance(TestCase):
    fixtures = ['sample_work_data', 'test_references']

    def test_display_title(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # short title from work if no alternate title
        assert la_vie.display_title() == la_vie.work.short_title
        # alternate title from instance if set
        la_vie.alternate_title = 'Family Life'
        assert la_vie.display_title() == la_vie.alternate_title

    def test_str(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        assert '%s (%s)' % (la_vie.display_title(), la_vie.copyright_year) \
            == str(la_vie)
        # no date
        la_vie.copyright_year = None
        assert '%s (n.d.)' % (la_vie.display_title(), )

    def test_generate_base_slug(self):
        work = Work.objects.create(primary_title='Ulysses')
        inst = Instance(work=work)
        # short title, no author or year
        assert inst.generate_base_slug() == 'ulysses'

        # single-name author
        joyce = Person.objects.create(authorized_name='Joyce')
        work.authors.add(joyce)
        assert inst.generate_base_slug() == 'joyce-ulysses'
        # comma-delimited author name
        joyce.authorized_name = 'Joyce, James'
        joyce.save()
        assert inst.generate_base_slug() == 'joyce-ulysses'

        # work year - used if no instance copyright year
        work.year = 1922
        work.save()
        assert inst.generate_base_slug() == 'joyce-ulysses-1922'
        # copyright year used when available
        inst.copyright_year = 1950
        assert inst.generate_base_slug() == 'joyce-ulysses-1950'

        # long titles truncated to ten words
        work.primary_title = 'A portrait of the artist as a young man: ' + \
            ' the strange story of Stephen Dedalus'
        work.save()
        assert inst.generate_base_slug() == 'joyce-a-portrait-of-the-artist-as-a-young-man-1950'

        # should handle unicode in titles
        emil_wk = Work.objects.create(primary_title="Émile ou de l'éducation")
        emil = Instance(work=emil_wk)
        assert emil.generate_base_slug() == 'emile-ou-de-leducation'
        # handle unicode in author names
        zizek = Person.objects.create(authorized_name='Žižek')
        emil_wk.authors.add(zizek)
        assert emil.generate_base_slug() == 'zizek-emile-ou-de-leducation'

    def test_generate_safe_slug(self):
        # should ignore itself when checking for duplicates
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        assert la_vie.generate_safe_slug() == la_vie.generate_base_slug()

        # new instance of the same edition, copy field not set, no other copies
        la_vie2 = Instance(work=la_vie.work, copyright_year=la_vie.copyright_year)
        assert la_vie2.generate_safe_slug() == '%s-B' % la_vie.generate_base_slug()
        # should also set copy value
        assert la_vie2.copy == 'B'

        # new instance of the same edition, with copy field set
        la_vie2 = Instance(work=la_vie.work, copyright_year=la_vie.copyright_year,
            copy='C')
        # should use copy field as set if it is unique
        assert la_vie2.generate_safe_slug() == \
            '%s-%s' % (la_vie.generate_base_slug(), la_vie2.copy)
        la_vie2.save()

        # create additional copy to test against multiple
        la_vie4 = Instance.objects.create(work=la_vie.work,
            copyright_year=la_vie.copyright_year, copy='D')
        # creating a new instance of the same edition with copy field unset
        la_vie3 = Instance(work=la_vie.work, copyright_year=la_vie.copyright_year)
        # should increment letter from the last copy letter encountered
        assert la_vie3.generate_safe_slug() == '%s-E' % la_vie.generate_base_slug()
        # should also set copy value
        assert la_vie3.copy == 'E'

    def test_save(self):
        # on save, if empty slug, should set one with generate safe slug
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        la_vie.slug = ''
        expected_slug = la_vie.generate_safe_slug()
        la_vie.save()
        la_vie.refresh_from_db()
        assert la_vie.slug == expected_slug

    def test_get_absolute_url(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        item_url = la_vie.get_absolute_url()
        resolved_url = resolve(item_url)
        assert resolved_url.url_name == 'detail'
        assert resolved_url.namespace == 'books'
        assert resolved_url.kwargs['slug'] == la_vie.slug

    def test_item_type(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        assert la_vie.item_type == 'Book'

        # journal article
        la_vie.journal = Journal.objects.all().first()
        assert la_vie.item_type == 'Journal Article'

        # book section
        la_vie.journal = None
        la_vie.collected_in = Instance()
        assert la_vie.item_type == 'Book Section'

    def test_author_names(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        assert la_vie.author_names() == la_vie.work.author_names()

    def test_catalogue_call_number(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # fixture has no call number (shelf-mark)
        assert la_vie.catalogue_call_numbers() == ''

        # add a first and second catalogue record
        owning_inst = OwningInstitution.objects.first()
        InstanceCatalogue.objects.create(institution=owning_inst,
            instance=la_vie, call_number='NY789', is_current=True)
        assert la_vie.catalogue_call_numbers() == 'NY789'
        InstanceCatalogue.objects.create(institution=owning_inst,
            instance=la_vie, call_number='PU456', is_current=True)
        assert la_vie.catalogue_call_numbers() == 'NY789, PU456'

    def test_clean(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        la_vie.journal = Journal.objects.all().first()
        la_vie.collected_in = Instance()
        with pytest.raises(ValidationError) as err:
            la_vie.clean()
        assert 'Cannot belong to both a journal and a collection' in str(err)

    def test_is_digitized(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        assert not la_vie.is_digitized()

        la_vie.digital_edition = Manifest()
        assert la_vie.is_digitized()

        # book section associated with a digitized work should be
        # considered digitized
        vie_part_wk = Work.objects.create()
        vie_part = Instance.objects.create(work=vie_part_wk)
        vie_part.collected_in = la_vie
        assert vie_part.is_digitized()

    def test_location(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # location is based on first part of manifest title
        la_vie.digital_edition = Manifest(
            label='Shelf - By the Cupboard - Title'
        )
        assert la_vie.location == 'Shelf, By the Cupboard'

        # some manifests have only one location, and second part
        # indicates it is a gift
        la_vie.digital_edition = Manifest(
            label='House - Gift Books, Works By and About Derrida, and Related Items - Derrida, Jacques. De la grammatologie.'
        )
        assert la_vie.location == 'House'

    def test_year(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # print year not known - use copyright year
        assert la_vie.year == la_vie.copyright_year
        # use print year if known
        la_vie.print_date = datetime(year=1965, month=1, day=1)
        assert la_vie.year == la_vie.print_date.year
        # nothing known
        la_vie.print_date_year_known = False
        la_vie.copyright_year = None
        assert la_vie.year is None

    def test_images(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # no digital edition - empty queryset
        assert la_vie.images().count() == 0
        # digital edition, no pages
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        assert la_vie.images().count() == 0
        Canvas.objects.bulk_create([
            Canvas(manifest=mfst, label='p1', order=1, short_id='c1'),
            Canvas(manifest=mfst, label='p2', order=2, short_id='c2'),
            Canvas(manifest=mfst, label='p3', order=3, short_id='c3'),
        ])
        assert la_vie.images().count() == 3
        assert Canvas.objects.first() in la_vie.images()

    def test_overview_images(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # no digital edition - empty queryset
        assert la_vie.overview_images().count() == 0
        # digital edition, no pages
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        assert la_vie.overview_images().count() == 0
        # create some test canvases to include/exclude
        outside_views = ['Front Cover', 'Inside Cover', 'Back Cover',
            'Spine', 'Edge View']
        for i, label in enumerate(outside_views):
            Canvas.objects.create(manifest=mfst, label=label,
                short_id='cover%d' % i, order=i)
        # normal page
        page = Canvas.objects.create(manifest=mfst, label='p. 33',
            short_id='page33', order=len(outside_views) + 1)
        # insertion
        insertion = Canvas.objects.create(manifest=mfst,
            label='pp. 33-34 Insertion A recto', short_id='insa',
            order=len(outside_views) + 2)

        overview_images = la_vie.overview_images()
        assert page not in overview_images
        assert insertion not in overview_images
        overview_labels = [c.label for c in overview_images]
        for label in outside_views:
            assert label in overview_labels

    def test_annotated_pages(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # no digital edition - empty queryset
        assert la_vie.annotated_pages().count() == 0
        # digital edition, no pages
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        assert la_vie.annotated_pages().count() == 0
        # normal pages, no annotations
        page = Canvas.objects.create(manifest=mfst, label='p. 33',
            short_id='page33', order=1)
        page2 = Canvas.objects.create(manifest=mfst, label='p. 35',
            short_id='page35', order=2)
        assert la_vie.annotated_pages().count() == 0
        # create an annotation
        Intervention.objects.create(canvas=page)
        assert la_vie.annotated_pages().count() == 1
        assert la_vie.annotated_pages().first() == page
        assert page2 not in la_vie.annotated_pages()
        # multiple annotations on the same page- should still only show once
        Intervention.objects.create(canvas=page)
        assert la_vie.annotated_pages().count() == 1

    def test_insertion_images(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        # no digital edition - empty queryset
        assert la_vie.insertion_images().count() == 0
        # digital edition, no pages
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        assert la_vie.insertion_images().count() == 0
        # cover
        cover = Canvas.objects.create(manifest=mfst, label='Front Cover',
            short_id='cov1', order=1)
        # normal page
        page = Canvas.objects.create(manifest=mfst, label='p. 33',
            short_id='page33', order=2)
        # insertion
        insertion = Canvas.objects.create(manifest=mfst,
            label='pp. 33-34 Insertion A recto', short_id='insa', order=3)
        insertion2 = Canvas.objects.create(manifest=mfst,
            label='pp. 33-34 Insertion A verso', short_id='insb', order=4)

        insertions = la_vie.insertion_images()
        assert page not in insertions
        assert cover not in insertions
        assert insertion in insertions
        assert insertion2 in insertions

    def test_allow_canvas_detail(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        # cover
        cover = Canvas.objects.create(manifest=mfst, label='Front Cover',
            short_id='cov1', order=1)
        # normal page
        page = Canvas.objects.create(manifest=mfst, label='p. 33',
            short_id='page33', order=2)
        # insertion
        insertion = Canvas.objects.create(manifest=mfst,
            label='pp. 33-34 Insertion A recto', short_id='insa', order=3)

        assert Instance.allow_canvas_detail(cover)
        assert not Instance.allow_canvas_detail(page)
        assert Instance.allow_canvas_detail(insertion)

        Intervention.objects.create(canvas=page)
        assert Instance.allow_canvas_detail(page)

    def test_allow_canvas_large_image(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        # cover
        cover = Canvas.objects.create(manifest=mfst, label='Front Cover',
            short_id='cov1', order=1)
        # normal page
        page = Canvas.objects.create(manifest=mfst, label='p. 33',
            short_id='page33', order=2)
        # insertion
        insertion = Canvas.objects.create(manifest=mfst,
            label='pp. 33-34 Insertion A recto', short_id='insa', order=3)

        # insertion and overview always allowed
        assert la_vie.allow_canvas_large_image(cover)
        assert la_vie.allow_canvas_large_image(insertion)
        # unannotated page not allowed
        assert not la_vie.allow_canvas_large_image(page)

        # annotated page allowed if not suppressed
        Intervention.objects.create(canvas=page)
        assert la_vie.allow_canvas_large_image(page)

        # suppress all
        la_vie.suppress_all_images = True
        assert not la_vie.allow_canvas_large_image(page)

        # suppress a different page
        la_vie.suppress_all_images = False
        la_vie.suppressed_images.add(cover)
        assert la_vie.allow_canvas_large_image(page)

        # suppress this specific page
        la_vie.suppressed_images.add(page)
        assert not la_vie.allow_canvas_large_image(page)

    def test_related_instances(self):
        # get la_vie and clone it
        la_vie = Instance.objects.filter(work__primary_title__icontains='la vie').first()
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m1')
        la_vie.copy = 'B'
        la_vie.save()
        pk = la_vie.pk
        la_vie.pk = None
        la_vie.digital_edition = mfst = Manifest.objects.create(short_id='m2')
        la_vie.slug = 'a-completely-different-slug'
        la_vie.copy = 'C'
        la_vie.save()
        # refresh the object so it has its pk and related objects
        la_vie.refresh_from_db()
        # original la_vie should be the only related instance
        assert len(la_vie.related_instances) == 1
        assert la_vie.related_instances[0].pk == pk

        # related instances for collected work should include any section authors
        vie_collection_wk = Work.objects.create()
        vie_collection = Instance.objects.create(work=vie_collection_wk)
        la_vie.collected_in = vie_collection
        la_vie.digital_edition = None
        la_vie.save()
        assert len(vie_collection.related_instances) == 1
        assert vie_collection.related_instances[0].pk == pk

        # delete digital edition
        la_vie_old = la_vie.related_instances[0]
        la_vie_old.digital_edition = None
        la_vie_old.save()
        # now check to get an empty set
        la_vie.refresh_from_db()
        assert len(la_vie.related_instances) == 0

    def test_primary_language(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")

        # no lang
        assert not la_vie.primary_language()

        lang_fr = Language.objects.get(name='French')
        lang_ger = Language.objects.get(name='German')
        lang_it = Language.objects.create(name='Italian')

        # one language on instance, not marked primary; one work language
        wklang = WorkLanguage.objects.create(language=lang_it, work=la_vie.work,
            is_primary=False)
        instlang = InstanceLanguage.objects.create(language=lang_fr, instance=la_vie,
            is_primary=False)
        assert la_vie.primary_language() == lang_fr

        # multiple languages, one marked primary
        instlang2 = InstanceLanguage.objects.create(language=lang_ger, instance=la_vie,
            is_primary=True)
        assert la_vie.primary_language() == lang_ger

        # work language used as fallback if no instance languages
        instlang.delete()
        instlang2.delete()
        # first work language used if only one
        assert la_vie.primary_language() == lang_it
        # primary work language used if set
        wklang2 = WorkLanguage.objects.create(language=lang_fr, work=la_vie.work,
            is_primary=True)
        # primary work language used
        assert la_vie.primary_language() == lang_fr

    def test_as_zotero_item(self):
        # mock a zotero library instance, since api call is required
        # to retrieve item template
        library = Mock()
        # return an empty dictionary for zotero template
        def get_template(*args, **kwargs):
            return dict()

        library.item_template.side_effect = get_template
        library.item_creator_types.return_value = [ # example creator types
            {'creatorType': 'author', 'localized': 'Author'},
            {'creatorType': 'editor', 'localized': 'Editor'},
            {'creatorType': 'translator', 'localized': 'Translator'}
        ]

        # test against each kind of item (book, journal article, book section)

        # ** book
        tropiques = Instance.objects.get(pk=134) # levi-strauss 'triste tropiques' 1955; book
        # add extra creators to the instances for testing
        editor = CreatorType.objects.get(name='Editor')
        translator = CreatorType.objects.get(name='Translator')
        bob = Person.objects.create(authorized_name='Smith, Bob')
        InstanceCreator.objects.create(instance=tropiques, creator_type=editor, person=bob)
        tropiques_z = tropiques.as_zotero_item(library)
        # test that the right item type template and creator types are retrieved
        library.item_template.assert_called_with('book')
        library.item_creator_types.assert_called_with('book')
        # check zotero item properties
        assert tropiques_z['url'] == tropiques.get_uri()
        assert tropiques_z['publisher'] == tropiques.publisher.name
        assert tropiques_z['place'] == tropiques.pub_place.first().name
        assert tropiques_z['language'] == 'fr'
        assert {
            'creatorType': 'editor',
            'firstName': 'Bob',
            'lastName': 'Smith'
        } in tropiques_z['creators']
        # check tags are set
        tags = [tag['tag'] for tag in tropiques_z['tags']]
        assert 'annotated' in tags
        assert 'extant' in tags
        assert 'translation' not in tags
        assert 'has dedication' not in tags
        assert 'has insertions' not in tags
        # assert {'tag': 'annotated'} in tropiques_z['tags']
        # assert {'tag': 'extant'} in tropiques_z['tags']
        assert tropiques_z['title'] == tropiques.work.primary_title
        assert tropiques_z['shortTitle'] == tropiques.work.short_title
        # cited in of grammatology but derrida work has no zotero id;
        # should not be not in collections
        assert not tropiques_z['collections']

        # set zotero id on "of grammatology" so it can be used as a collection
        grammatology = DerridaWork.objects.get(pk=1)
        grammatology.zotero_id = 'ABCDEF'
        grammatology.save()

        # ** book section
        library.reset_mock()
        lecriture = Instance.objects.get(pk=92) # cohen 'la grande invention de lecriture' 1958; book section
        InstanceCreator.objects.create(instance=lecriture, creator_type=translator, person=bob)
        lecriture_z = lecriture.as_zotero_item(library)
        # test that the right item type template and creator types are retrieved
        # book section also uses book to get book-level metadata
        library.item_template.assert_any_call('bookSection')
        library.item_template.assert_any_call('book')
        library.item_creator_types.assert_any_call('bookSection')
        library.item_creator_types.assert_any_call('book')
        assert lecriture_z['title'] == lecriture.work.primary_title
        assert lecriture_z['shortTitle'] == lecriture.work.short_title
        assert lecriture_z['bookTitle'] == lecriture.collected_in.display_title()
        assert lecriture_z['date'] == 1958
        assert {
            'creatorType': 'translator',
            'firstName': 'Bob',
            'lastName': 'Smith'
        } in lecriture_z['creators']
        # check tags are set
        tags = [tag['tag'] for tag in lecriture_z['tags']]
        assert 'extant' in tags
        assert 'digital edition' not in tags
        # cited in 'of grammatology'
        assert grammatology.zotero_id in lecriture_z['collections']

        # ** journal article
        lemot = Instance.objects.get(pk=166) # martinet 'le mot' 1965; journal article in 'diogene'
        lemot_z = lemot.as_zotero_item(library)
        # test that the right item type template and creator types are retrieved
        library.item_template.assert_called_with('journalArticle')
        library.item_creator_types.assert_called_with('journalArticle')
        # check that item-specific properties are present
        assert lemot_z['publicationTitle'] == "Diog\u00e8ne"
        assert lemot_z['pages'] == '39-53'
        # place not allowed for journal article
        assert 'place' not in lemot_z
        tags = [tag['tag'] for tag in lemot_z['tags']]
        assert 'extant' not in tags
        assert 'annotated' not in tags
        # check that creator is present
        assert {
            'creatorType': 'author',
            'firstName': 'Andr\u00e9',
            'lastName': 'Martinet',
        } in lemot_z['creators']
        # cited in 'of grammatology'
        assert grammatology.zotero_id in lemot_z['collections']

        # item with existing zotero id
        lemot.zotero_id = 'Z12345'
        zotero_item = lemot.as_zotero_item(library)
        assert zotero_item['key'] == lemot.zotero_id

        # item with catalogue / owning institution
        instance = Instance.objects.filter(owning_institutions__isnull=False).first()
        zotero_item = instance.as_zotero_item(library)
        assert zotero_item['archive'] == instance.owning_institutions.first().name

        # item with PUL finding aid URL
        instance = Instance.objects.filter(uri__contains='findingaids').first()
        zotero_item = instance.as_zotero_item(library)
        assert zotero_item['archiveLocation'] == instance.uri

        # check notes in abstract
        # - should include number of references
        instances_with_refs = Instance.objects.annotate(ref_count=Count('reference'))
        # get an instance with a single reference
        instance = instances_with_refs.filter(ref_count=1).first()
        # set copy information to test inclusion in notes
        instance.copy = 'B'
        zotero_item = instance.as_zotero_item(library)
        assert 'Copy {}'.format(instance.copy) in zotero_item['abstractNote']
        assert '1 reference' in zotero_item['abstractNote']

        # get an instance with more than one reference
        instance = instances_with_refs.filter(ref_count__gt=1).first()
        # set copy information to test inclusion in notes
        zotero_item = instance.as_zotero_item(library)
        assert '{} references'.format(instance.reference_set.count()) \
            in zotero_item['abstractNote']

        # instance with annotations documented - not currently represented
        # in fixture data


class TestInstanceQuerySet(TestCase):
    fixtures = ['sample_work_data.json']

    def setUp(self):
        self.manif1 = Manifest.objects.create(short_id='bk123', label='Foobar')

    def test_with_digital_eds(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")

        # no digital editiosn with manifests, so should return empty queryset
        empty = Instance.objects.with_digital_eds()
        assert len(empty) == 0

        # associate la_vie with the set up manfest, len should be 1
        # and the only result should be la_vie
        la_vie.digital_edition = self.manif1
        la_vie.save()

        qs = Instance.objects.with_digital_eds()
        assert len(qs) == 1
        assert qs[0] == la_vie

class TestWorkLanguage(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        la_vie = Work.objects.get(short_title__contains="La vie")
        fr = Language.objects.get(name='French')
        wklang = WorkLanguage(language=fr, work=la_vie)
        assert str(wklang) == '%s %s' % (la_vie, fr)

        wklang.is_primary = True
        assert str(wklang) == '%s %s (primary)' % (la_vie, fr)


class TestWorkSubject(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        la_vie = Work.objects.get(short_title__contains="La vie")
        ling = Subject(name='Linguistics')
        wksubj = WorkSubject(work=la_vie, subject=ling)
        assert str(wksubj) == '%s %s' % (la_vie, ling)

        wksubj.is_primary = True
        assert str(wksubj) == '%s %s (primary)' % (la_vie, ling)

    def test_work_count(self):
        # test abstract work count mix-in via owning subject model
        # tests that html for admin form is rendered correctly
        la_vie = Work.objects.get(short_title__contains="La vie")
        subj = Subject.objects.create(name='Linguistics')
        # new subject has no books associated
        change_url = reverse('admin:books_work_changelist')
        admin_work_count = subj.work_count()
        assert change_url in admin_work_count
        assert 'id__exact=%s' % la_vie.pk in admin_work_count
        assert '0' in admin_work_count

        # add a work/subject
        WorkSubject.objects.create(work=la_vie, subject=subj)
        admin_work_count = subj.work_count()
        assert '1' in admin_work_count


class TestInstanceLanguage(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")
        ger = Language.objects.get(name='German')
        instlang = InstanceLanguage(language=ger, instance=la_vie)
        assert str(instlang) == '%s %s' % (la_vie, ger)

        instlang.is_primary = True
        assert str(instlang) == '%s %s (primary)' % (la_vie, ger)


class TestInstanceCatalogue(TestCase):
    fixtures = ['sample_work_data.json']

    def test_str(self):
        la_vie = Instance.objects.get(work__short_title__contains="La vie")

        inst_cat = la_vie.instancecatalogue_set.first()
        assert str(inst_cat) == '%s / %s' % (la_vie, inst_cat.institution)

        # with dates
        inst_cat.start_year = 1891
        assert str(inst_cat) == '%s / %s (1891-)' % (la_vie, inst_cat.institution)
